import arzumify.polyenergy.api.EnergyProvider;
import arzumify.polyenergy.api.EnergyReceiver;
import arzumify.polyenergy.impl.CoordinateMatchMaker;
import arzumify.presence.presences.IntegerPresence;
import arzumify.presence.presences.Presence;
import net.fabricmc.fabric.api.event.lifecycle.v1.ServerBlockEntityEvents;
import net.minecraft.block.entity.BlockEntity;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.math.Vec3i;

import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

// Cables can have ONE provider and MANY receivers.
public class SimpleCodeOnlyCable implements EnergyProvider, EnergyReceiver, ServerBlockEntityEvents.Unload, ServerBlockEntityEvents.Load {
    public final long capacity;
    public final long inputRate;
    public final long outputRate;
    public final IntegerPresence presence;
    public final String name;
    public long energy = 0;
    public final Set<EnergyProvider> providers = ConcurrentHashMap.newKeySet();

    public SimpleCodeOnlyCable(Vec3i pos, long capacity, long inputRate, long outputRate, String name) {
        this.capacity = capacity;
        this.inputRate = inputRate;
        this.outputRate = outputRate;
        this.name = name;
        this.presence = new IntegerPresence((short) 1, new arzumify.presence.maths.Vec3i(pos.getX(), pos.getY(), pos.getZ()));
        CoordinateMatchMaker.addReceiver(this);
    }

    public static void tick(SimpleCodeOnlyCable cable) {
        System.out.println(cable.name + " ticking");
        var leftToFill = Math.min(cable.inputRate, cable.capacity - cable.energy);
        for (EnergyProvider provider : cable.providers) {
            // Less than zero shouldn't be possible, but just in case...
            if (leftToFill <= 0) {
                break;
            }
            long extracted = provider.extract(leftToFill, cable);
            cable.energy += extracted;
            leftToFill -= extracted;
        }
    }

    @Override
    public long extract(long amount, EnergyReceiver receiver) {
        System.out.println(name + " extracting " + amount + " energy");
        long extracted = Math.min(Math.min(outputRate, amount), energy);
        energy -= extracted;
        System.out.println(name + " extracted " + extracted + " energy");
        return extracted;
    }

    @Override
    public void exists(EnergyReceiver receiver) {
        System.out.println(name + " found other receiver");
        if (!providers.isEmpty()) {
            receiver.ready(this);
        }
    }

    @Override
    public void ready(EnergyProvider provider) {
        System.out.println(name + " found other provider");
        providers.add(provider);
        if (providers.size() == 1) {
            CoordinateMatchMaker.addProvider(this);
        }
    }

    @Override
    public void unready(EnergyProvider provider) {
        providers.remove(provider);
        if (providers.isEmpty()) {
            CoordinateMatchMaker.removeProvider(this);
        }
    }

    @Override
    public void onLoad(BlockEntity blockEntity, ServerWorld serverWorld) {
        CoordinateMatchMaker.addReceiver(this);
    }

    @Override
    public void onUnload(BlockEntity blockEntity, ServerWorld serverWorld) {
        CoordinateMatchMaker.removeProvider(this);
        CoordinateMatchMaker.removeReceiver(this);
        providers.clear();
    }

    @Override
    public Presence<IntegerPresence> presence() {
        return presence;
    }
}
